// Program associated with the paper entitled "N
// "Nonlinear bounded-error state estimation of continuous-time systems"
// submitted to Automatica
// This file contains the algorithms
//---------------------------------------------------------------------------
#include <vcl.h>
#pragma hdrstop
#include <stdlib.h>
#include "algo.h"
#include "repere.h"
#include "box.h"
#define dimx 1000   //size of the arrays X[k] and Y[k]
double Te;   //sampling time
int kmax;    //number of samples
//---------------------------------------------------------------------------
#pragma package(smart_init)
double epsilon;  //accuracy
repere R;
box tube[dimx+2];       // array containing enclosure boxes [x](k) for the state vector x(k)
box goulot[dimx+2];
box W[dimx+2];
box Xoo(2,true);        //  large box containing the states of interest
double ay=1;            //  noise
double by=0.5;          //  pessimism on the noise
interval Y[dimx+2];     //  measurement intervals
box Xempty(2,true);     //  empty box
//-------------------------------------------------------------------------------------
void Init()
{   randomize();
    Xoo[1]=interval(0,100);  Xoo[2]=interval(0,200);
    Xempty.isEmpty=true;
    for (int k=0;k<=kmax+1;k++)  {tube[k]=Xoo;}   //i.e. nothing is known on the state vector
}
//-------------------------------------------------------------------------------------
void Fd(box& R, box& X, box& W, double T)     // Discretization of the evolution function
{   R[1]=Inter(R[1],X[1]+T*(1-0.01*W[2])*W[1]);
    R[2]=Inter(R[2],X[2]+T*(-1+0.02*W[1])*W[2]);
    if (R[1].isEmpty||R[2].isEmpty) R.isEmpty=true;
}
//-------------------------------------------------------------------------------------
void SIMUL()         //Generation of the data by simulation of the voltera differential
{   box X(2,true);   //equation with initial condition with x1(0)=x2(0)=50
    box Xp(2,true);
    X[1]=50;X[2]=50;
    for (int i=0;i<=(kmax+1)*100;i++)
    {   if ((i%100)==0)
        { int k=i/100;
          R.DrawCadre(X,clWhite,bsSolid,2);
          Y[k]=Center(X[1])+Pioche(interval(-ay,ay))+interval(-ay-by,ay+by);
        }
        Xp=Xoo;
        Fd(Xp,X,X,Te/100);  //To limit the inaccuracy of Euler's method, the sampling time is taken as Te/100
        X=Xp;
    }
} ;
//-------------------------------------------------------------------------------------
int Contract(box* X)    // See the paper
{   for (int k=0;k<=kmax;k++)
    {   box Xh=Xoo;
        Fd(Xh,X[k],X[k],Te);
        W[k]=Union(X[k],Xh);  W[k]=Inflate(W[k],0.1*Width(W[k])+0.0001);
        Fd(X[k+1],X[k],W[k],Te);
        if (!Subset(X[k+1],W[k]))
            {R.DrawCadre(Xoo,clBlue,bsSolid); return (-1);} //Error.
    }
    for (int k=kmax;k>=0;k--)
    {   Fd(X[k],X[k+1],W[k],-Te);
        X[k][1]= Inter(X[k][1],Y[k]);
        if (X[k][1].isEmpty || X[k][2].isEmpty ) {X[k].isEmpty=true; return 0;};
    }
    return 1;
};
//-------------------------------------------------------------------------------------
void SIVIA(box& Xk0, int k0)     // recursive algorithm corresponding to Steps 2 to 6 of
{   box X[dimx+2];               // the algorithm STRANGLE in the paper
    for (int k=0;k<=kmax+1;k++) {X[k]=tube[k];}   // See Step 3 of the STRANGLE in the paper
    X[k0]=Xk0;
    interval Xk01=X[k0][1];
    interval Xk02=X[k0][2];
    if (Contract(X)==0) return;
    Xk0=X[k0];
    if (Width(Xk0)<epsilon)
    {   R.DrawBox(Xk0,clRed,clPurple,bsSolid);
        for (int k=0;k<=kmax+1;k++) {goulot[k]=Union(X[k],goulot[k]);}
        return;
    };
    box X1(2,true);   box X2(2,true);
    Bisect(Xk0,X1,X2);  SIVIA(X1,k0);    SIVIA(X2,k0);
    Xk0=Union(X1,X2);
} ;
//-------------------------------------------------------------------------------------
void Strangle (int k0)          // see the paper
{   for (int k=0;k<=kmax;k++) {goulot[k]=Xempty;}
    SIVIA(tube[k0],k0);
    for (int k=0;k<=kmax;k++) {tube[k]=Inter(tube[k],goulot[k]);}
}
//-------------------------------------------------------------------------------------
void StateBound(repere& R1)     //Measurements are available
{   Te=0.03;  kmax=200;         //See Figure 2 of the paper
    R=R1;  Init();
    SIMUL();   //Generate the data
    for (epsilon=40;epsilon>2;epsilon=epsilon/4)
    {   for (int i=1;i<=30;i++) Strangle(rand()%(kmax));
        for (int k=0;k<=kmax;k++)  {R.DrawBox(tube[k],clGreen,clBlack,bsSolid);};
    };
    SIMUL();  // Draw the white curves of Figure 2
}
//-------------------------------------------------------------------------------------
void Simulation(repere& R1)    // prediction (no measurements)
{   Te=0.005; kmax=400;        // see Figure 3 of the paper
    R=R1;  Init();
    tube[0][1]=interval(49.5,50.5); tube[0][2]=interval(49.5,50.5);
    for (int k=0;k<=kmax;k++)   {Y[k]=interval(-1000,1000);  }
    for (epsilon=40;epsilon>0.5;epsilon=epsilon/4)
    {   Strangle(0);
        for (int i=1;i<=10;i++)    {int k0=rand()%(kmax+1); Strangle(k0);}
        for (int k=0;k<=kmax;k++)  {R.DrawBox(tube[k],clGreen,clBlack,bsSolid);};
    }
    SIMUL();  // Draw the white curve of Figure 3
} ;
//-------------------------------------------------------------------------------------

